/*
 * GNOME hardware sensors panel applet
 * (C) 1999 The Free Software Foundation
 *
 * Displays the state of hardware sensors in your machine.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *                                
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *           
 * Authors: 	Brendon Humphrey (brendy@swipnet.se)
 *
 * This applet is based on the diskusage applet and the 'sensors'
 * application that is included in the lm_sensors package.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>

#include <sensors/sensors.h>
#include <sensors/chips.h>
#include "chips.h"

#define MAX_NUM_CHIPS 20

sensors_chip_name chips[MAX_NUM_CHIPS];
int               chips_count = 0;

/* ==========================================================================
 * add_sensors_data
 * =========================================================================*/
void add_sensors_data( 
	machine *all_sensors, 
	char alarm, 
	char *name, 
	double min, 
	double max, 
	double now,
	sensor_units units)
{
	all_sensors->sensors = realloc(all_sensors->sensors, (1+all_sensors->num_sensors)*sizeof(sensor_reading));
	
	all_sensors->sensors[all_sensors->num_sensors].alarm = alarm;
	strcpy( all_sensors->sensors[all_sensors->num_sensors].name, name);
	all_sensors->sensors[all_sensors->num_sensors].min = min;
	all_sensors->sensors[all_sensors->num_sensors].max = max;
	all_sensors->sensors[all_sensors->num_sensors].now = now;
	all_sensors->sensors[all_sensors->num_sensors].units = units;
	all_sensors->num_sensors++;
}

/* ==========================================================================
 * free_the_label
 * =========================================================================*/
void free_the_label( char **label)
{
	if (*label)
		free(*label);
	*label = NULL;
}

/* ==========================================================================
 * get_sensors
 * =========================================================================*/
void get_sensors( 
	machine *all_sensors, 
	int first, 
	int last, 
	const sensors_chip_name *name, 
	sensor_units units)
{
	char *label = NULL;
	double cur=0,min=0,max=0;
	int i;
	
	for(i=first;i<=last;i++)
	{
		if (!sensors_get_label(*name,i,&label) &&
				!sensors_get_feature(*name,i,&cur)) 
		{
			add_sensors_data(all_sensors, 'n', label, min, max, cur, units);
		} else
			printf("ERROR: Can't get feature %i from name %s\n", i, name->prefix);
		free_the_label(&label);
	}
}

/* ==========================================================================
 * read_lm75
 * =========================================================================*/
void read_lm75(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_LM75_TEMP, SENSORS_LM75_TEMP, name, CELSIUS);
}

/* ==========================================================================
 * read_adm1021
 * =========================================================================*/
void read_adm1021(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_ADM1021_TEMP, SENSORS_ADM1021_TEMP, name, CELSIUS);
	get_sensors( all_sensors, SENSORS_ADM1021_REMOTE_TEMP, SENSORS_ADM1021_REMOTE_TEMP, name, CELSIUS);	
}

/* ==========================================================================
 * read_adm9240
 * =========================================================================*/
void read_adm9240(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_ADM9240_IN0, SENSORS_ADM9240_IN5, name, VOLTS);
	get_sensors( all_sensors, SENSORS_ADM9240_FAN1, SENSORS_ADM9240_FAN2, name, RPM);	
	get_sensors( all_sensors, SENSORS_ADM9240_TEMP, SENSORS_ADM9240_TEMP, name, CELSIUS);	
	get_sensors( all_sensors, SENSORS_ADM9240_VID, SENSORS_ADM9240_VID, name, VOLTS);	
}

/* ==========================================================================
 * read_sis5595
 * =========================================================================*/
void read_sis5595(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_SIS5595_IN0, SENSORS_SIS5595_IN3, name, VOLTS);
	get_sensors( all_sensors, SENSORS_SIS5595_FAN1, SENSORS_SIS5595_FAN2, name, RPM);	
	get_sensors( all_sensors, SENSORS_SIS5595_TEMP, SENSORS_SIS5595_TEMP, name, CELSIUS);	
}

/* ==========================================================================
 * read_lm78
 * =========================================================================*/
void read_lm78(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_LM78_IN0, SENSORS_LM78_IN6, name, VOLTS);
	get_sensors( all_sensors, SENSORS_LM78_FAN1, SENSORS_LM78_FAN3, name, RPM);	
	get_sensors( all_sensors, SENSORS_LM78_TEMP, SENSORS_LM78_TEMP, name, CELSIUS);	
	get_sensors( all_sensors, SENSORS_LM78_VID, SENSORS_LM78_VID, name, VOLTS);	
}

/* ==========================================================================
 * read_gl518
 * =========================================================================*/
void read_gl518(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_GL518_VDD, SENSORS_GL518_VDD, name, VOLTS);
	get_sensors( all_sensors, SENSORS_GL518_VIN1, SENSORS_GL518_VIN3, name, VOLTS);	
	get_sensors( all_sensors, SENSORS_GL518_FAN1, SENSORS_GL518_FAN2, name, RPM);	
	get_sensors( all_sensors, SENSORS_GL518_TEMP, SENSORS_GL518_TEMP, name, CELSIUS);	
} 

/* ==========================================================================
 * read_gl520
 * =========================================================================*/
void read_gl520(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_GL520_VDD, SENSORS_GL520_VIN4, name, VOLTS);
	get_sensors( all_sensors, SENSORS_GL520_FAN1, SENSORS_GL520_FAN2, name, RPM);	
	get_sensors( all_sensors, SENSORS_GL520_TEMP1, SENSORS_GL520_TEMP2, name, CELSIUS);
	get_sensors( all_sensors, SENSORS_GL520_VID, SENSORS_GL520_VID, name, VOLTS);	
}

/* ==========================================================================
 * read_lm80
 * =========================================================================*/
void read_lm80(machine *all_sensors, const sensors_chip_name *name)
{	
	get_sensors( all_sensors, SENSORS_LM80_IN0, SENSORS_LM80_IN6, name, VOLTS);
	get_sensors( all_sensors, SENSORS_LM80_FAN1, SENSORS_LM80_FAN2, name, RPM);	
	get_sensors( all_sensors, SENSORS_LM80_TEMP, SENSORS_LM80_TEMP, name, CELSIUS);	
}

/* ==========================================================================
 * read_w83781d
 * =========================================================================*/
void read_w83781d(machine *all_sensors, const sensors_chip_name *name)
{
	int is82d;
	int is83s;
	
	is82d = (!strcmp(name->prefix,"w83782d")) ||
		(!strcmp(name->prefix,"w83627hf"));
	is83s = !strcmp(name->prefix,"w83783s");
	
	get_sensors( all_sensors, SENSORS_W83781D_IN0, SENSORS_W83781D_IN6, name, VOLTS);
	
	if(is82d)
	{
		get_sensors( all_sensors, SENSORS_W83782D_IN7, SENSORS_W83782D_IN8, name, VOLTS);
	}
	get_sensors( all_sensors, SENSORS_W83781D_FAN1, SENSORS_W83781D_FAN3, name, RPM);
	get_sensors( all_sensors, SENSORS_W83781D_TEMP1, SENSORS_W83781D_TEMP1, name, CELSIUS);
	get_sensors( all_sensors, SENSORS_W83781D_TEMP2, SENSORS_W83781D_TEMP2, name, CELSIUS);
	
	if(!is83s)
	{
		get_sensors( all_sensors, SENSORS_W83781D_TEMP3, SENSORS_W83781D_TEMP3, name, CELSIUS);
	}	
	
	get_sensors( all_sensors, SENSORS_W83781D_VID, SENSORS_W83781D_VID, name, VOLTS);	
}

/* ==========================================================================
 * read_maxilife
 * =========================================================================*/
void read_maxilife(machine *all_sensors, const sensors_chip_name *name)
{
	get_sensors( all_sensors, SENSORS_MAXI_CG_TEMP1, SENSORS_MAXI_CG_TEMP5, name, CELSIUS);	
	get_sensors( all_sensors, SENSORS_MAXI_CG_FAN1, SENSORS_MAXI_CG_FAN3, name, RPM);	
	get_sensors( all_sensors, SENSORS_MAXI_CG_PLL, SENSORS_MAXI_CG_PLL, name, MHZ);	
	get_sensors( all_sensors, SENSORS_MAXI_CG_VID1, SENSORS_MAXI_CG_VID4, name, VOLTS);	
}

/* ==========================================================================
 * fetch_sensor_for_display
 * =========================================================================*/
void fetch_sensor_for_display( machine *all_sensors,int index, char *name, char *value, char *units)
{
	/* these are format string for various sensor readout types */
	static char *sensor_fmt_strings[] = {"%-.1f", "%.0f", "%.0f", "%.0f", "%.1f"};
	static char *sensor_unit_labels[] = {"Volts", "Celsius", "F'heit", "RPM", "Mhz"};
	
	strcpy(name, all_sensors->sensors[index].name); 
	sprintf(value, sensor_fmt_strings[all_sensors->sensors[index].units], all_sensors->sensors[index].now);
	strcpy(units, sensor_unit_labels[all_sensors->sensors[index].units]);
}

/* ==========================================================================
 * free_sensors
 * =========================================================================*/
void free_sensors( machine *all_sensors)
{
	if (all_sensors->num_sensors > 0)
	{
		all_sensors->num_sensors=0;
		free(all_sensors->sensors);
		all_sensors->sensors=NULL;
	}
}

/* ==========================================================================
 * identify_chip
 * =========================================================================*/
int identify_chip( char *name, known_chips *chip)
{
	int	found = 1;
	
	if (!strcmp(name,"lm75"))
		*chip=LM75;
	else if (!strcmp(name,"adm1021") || !strcmp(name,"max1617") ||
			!strcmp(name,"max1617a") || !strcmp(name, "thmc10"))
		*chip=ADM1021;
	else if (!strcmp(name,"adm9240") ||
			!strcmp(name,"ds1780") ||
			!strcmp(name,"lm81"))
		*chip=ADM9240;
	else if (!strcmp(name,"lm78") || !strcmp(name,"lm78-j") ||
			!strcmp(name,"lm79") || !strcmp(name,"sis5595"))
		*chip = LM78;
	else if (!strcmp(name,"sis5595"))
		*chip = SIS5595;
	else if (!strcmp(name,"lm80"))
		*chip = LM80;
	else if (!strcmp(name,"gl518sm"))
		*chip = GL518;
	else if (!strcmp(name, "gl520sm"))
		*chip = GL520;
	else if ((!strcmp(name,"w83781d")) ||
			(!strcmp(name,"w83782d")) ||
			(!strcmp(name,"w83783s")) ||
			(!strcmp(name,"w83627hf")))
		*chip = W83781D;
	else if (!strcmp(name,"maxilife-cg") ||
			!strcmp(name,"maxilife-co") ||
			!strcmp(name,"maxilife-as"))
		*chip=MAXILIFE;
	else
	{
		found = 0;
		printf("unknown chip encountered:%s\n",name);
	}
	
	return found;
}

/* ==========================================================================
 * read_sensors
 * =========================================================================*/
void read_sensors(machine *all_sensors, sensors_chip_name name)
{
	const char *algo;
	const char *adap;
	known_chips chip;
	
	/* extract chip and algo - not needed in final version */
	adap = sensors_get_adapter_name(name.bus);
	
	if (adap)
		printf("Adapter: %s\n",adap);
	
	algo = sensors_get_algorithm_name(name.bus);
	if (algo)
		printf("Algorithm: %s\n",algo);
	
	if (!algo || !adap)
		printf(" ERROR: Can't get adapter or algorithm?!?\n");
	
	/* identify the chip involved */
	if( identify_chip(name.prefix, &chip) )
	{
		switch(chip)
		{
			case W83781D:
				read_w83781d(all_sensors, &name);
				break;
			case MAXILIFE:
				read_maxilife(all_sensors, &name);
				break;
			case LM80:
				read_lm80(all_sensors, &name);
				break;
			case GL518:
				read_gl518(all_sensors, &name);
				break;
			case GL520:
				read_gl520(all_sensors, &name);
				break;
			case LM78:
				read_lm78(all_sensors, &name);
				break;
			case SIS5595:
				read_sis5595(all_sensors, &name);
				break;
			case ADM9240:
				read_adm9240(all_sensors, &name);
				break;
			case ADM1021:
				read_adm1021(all_sensors, &name);
				break;
			case LM75:
				read_lm75(all_sensors, &name);
				
			default:
				break;
		}
	}
}

/* ==========================================================================
 * get_sensor_data
 * =========================================================================*/
void get_sensor_data( machine *all_sensors )
{
	const sensors_chip_name *chip;
	int                     chip_nr;
	int                     i;
	
	/* always free the memory allocated by the last pass */
	/* through the sensors */
	free_sensors(all_sensors);

	/* iterate through all of the hardware on the machine */	
	for (chip_nr = 0; (chip = sensors_get_detected_chips(&chip_nr));)
	{
		for(i = 0; i < chips_count; i++)
		{			
			if (sensors_match_chip(*chip,chips[i])) {
				read_sensors(all_sensors, *chip);
				i = chips_count;
			}
		}
	}	
}

/* ==========================================================================
 * initialise_chips
 * =========================================================================*/
void initialise_chips()
{
	chips[0].prefix = SENSORS_CHIP_NAME_PREFIX_ANY;
	chips[0].bus = SENSORS_CHIP_NAME_BUS_ANY;
	chips[0].addr = SENSORS_CHIP_NAME_ADDR_ANY;
	chips_count = 1;
}
